require "numru/gphys"
require "no_rdb_base"

class AnalysisColumn < ActiveRecord::ConnectionAdapters::Column

  def type_cast_code(var_name)
    case type
    when :array
      "#{self.class.name}.value_to_array(#{var_name})"
    when :array_int
      "#{self.class.name}.value_to_array(#{var_name}, :int)"
    when :array_float
      "#{self.class.name}.value_to_array(#{var_name}, :float)"
    when :model
      "#{self.class.name}.check_model(#{var_name})"
    else
      super
    end
  end

  def self.value_to_array(value, val_type=nil)
    if Array === value
      ary = value
    elsif Hash === value
      ary = Array.new
      value.each{|k,v|
        ary[k.to_i] = v
      }
    elsif String === value
      ary = value.split(/,/)
    else
      ary = [value]
    end
    case val_type
    when :int
      ary.collect!{|v| v.to_i}
    when :float
      ary.collect!{|v| v.to_f}
    end
    return ary
  end

  def self.check_model(value)
    if ActiveRecord::Base === value
      return value
    else
      raise "a model is expected"
    end
  end

  def default
    Array===@default ? @default.dup : @default
  end
    

  private
  def simplified_type(type)
    case type
    when /array_int/i
      :array_int
    when /array_float/i
      :array_float
    when /array/i
      :array
    when /model/i
      :model
    else
      super(type)
    end
  end

end

class Analysis < ActiveRecord::NoRdbBase

  ACTION_TYPE = %w( draw analysis )

  DRAW_TYPE = %w( line tone vector )

  DRAW_PROJECTION = {1 => "rectangular uniform coordinate",
                     2 => "semi-logarithmic coordinate (y axis)",
                     3 => "semi-logarithmic coordinate (x axis)",
                     4 => "logarithmic coordinate",
                     5 => "polar coordinate",
                     6 => "bipolar coordinate",
#                     7 => "elliptic coordinate",
                     10 => "equidistant cylindrical projection",
                     11 => "Mercator's projection",
                     12 => "Mollweide's projection",
                     13 => "Hammer's projection",
                     14 => "Eckert VI projection",
                     15 => "Kitada's elliptic projection",
                     20 => "equidistant conical projection",
                     21 => "Lambert's equal-area conical projection",
                     22 => "Lambert's conformal conical projection",
                     23 => "Bonne's projection",
                     30 => "orthographic projection",
                     31 => "polar stereo projection",
                     32 => "azimuthal equidistant projection",
                     33 => "Lambert's azimuthal equal-area projection"
                    }

  DRAW_SIZE = [[700,700], [550,550], [400,400], [250,250]]

  common_attrs = [
     ["variables", [], "array", false],
     ["action_type", ACTION_TYPE.index("draw"), "int", false]
  ]
  draw_common_attrs = [
    ["draw_share", true, "boolean", false],
    ["draw_type", DRAW_TYPE.index("tone"), "int", false],
    ["x_axis", nil, "string", false],
    ["y_axis", nil, "string"],
    ["draw_projection", 1, "int", false],
    ["region", {}, "hash", false],
    ["draw_keep", false, "boolean", false],
    ["draw_size", 1, "int", false],
    ["anim", false, "boolean"],
    ["anim_dim", nil, "string"],
    ["viewport", "0.2, 0.8, 0.2, 0.8", "array_float"]
  ]
  @@draw_line_attrs = [
  ]
  @@draw_tone_attrs = [
    ["tone", true, "boolean", false, :vizshot],
    ["contour", true, "boolean", false, :vizshot],
    ["min", nil, "float", true, :ggraph],
    ["max", nil, "float", true, :ggraph],
    ["nlev", nil, "int", true, :ggraph],
    ["levels", nil, "array_float", true, :ggraph],
    ["color_bar", true, "boolean", false, :vizshot],
    ["log", false, "boolean", false, :ggraph]
  ]
  @@draw_vector_attrs = [
    ["vector_variables_order", [], "array", false]
  ]
  function_attrs = [
    ["function", nil, "model", false],
    ["input_num", nil, "int", false],
    ["function_arguments", [], "array", false],
    ["function_variables_order", [], "array", false]
  ]

  @@minmaxs = {
    "action_type" => [0, ACTION_TYPE.length-1],
    "draw_type" => [0, DRAW_TYPE.length-1],
    "draw_size" => [0, DRAW_SIZE.length-1],
    "input_num" => [1, Float::MAX]
  }



  (common_attrs +
   draw_common_attrs +
   @@draw_line_attrs + @@draw_tone_attrs + @@draw_vector_attrs +
   function_attrs
  ).each{|ary|
    push_column AnalysisColumn.new( *ary[0..3] )
  }

  def write_attribute(attr_name, value)
    case attr_name.to_s
    when "function"
      unless Function === value
        value = Function.find(:first, :conditions=>["id=?",value]) || Function.find(:first, :conditions=>["name=?",value])
      end
    when "vector_variables_order"
      value = variables_order(value)
    when "function_variables_order"
      value = variables_order(value)
    end

    super

    if (minmax = @@minmaxs[attr_name.to_s])
      val = read_attribute(attr_name)
      if val < minmax[0]
        write_attribute(attr_name, minmax[0].to_s)
      elsif val > minmax[1]
        write_attribute(attr_name, minmax[1].to_s)
      end
    end
    xa = read_attribute("x_axis")
    ya = read_attribute("y_axis")
    ad = read_attribute("anim") && read_attribute("anim_dim")
    case attr_name.to_s
    when "x_axis"
      if xa
        if xa == ya
          write_attribute("y_axis", nil)
        elsif xa == ad
          write_attribute("anim_dim", nil)
        end
      end
    when "y_axis"
      if ya
        if ya == xa
          write_attribute("x_axis", nil)
        elsif ya == ad
          write_attribute("anim_dim", nil)
        end
      end
    when "anim_dim"
      if ad
        if ad == xa
          write_attribute("x_axis", nil)
        elsif ad == ya
          write_attribute("y_axis", nil)
        end
      end
    when "draw_projection"
      unless DRAW_PROJECTION[read_attribute("draw_projection")]
        write_attribute("draw_projection", "1")
      end
    end
  end

  attr_accessor :user

  def variables
    vars = read_attribute("variables")
    vars.each{|var|
      next if var.new_record?
      unless var == Variable.find_for_user(@user, :first, :conditions=>["id=?",var.id])
        raise "an invalide variable was set"
      end
    }
    return vars
  end
  def function
    func = read_attribute("function")
    return func if func.nil?
    unless func ==  Function.find_for_user(@user, :first, :conditions=>["id=?",func.id])
      raise "an invalide function was set"
    end
    return func
  end


  def variable_clear
    @dimensions = nil
    %w( variables region x_axis y_axis vector_variables_order function_variables_order ).each{|name|
      column = column_for_attribute(name)
      default = column.default
      self[name] = default
    }
  end

  def file_and_variable_names
    variables.collect do |var|
      var.file_and_variable_names
    end
  end

  def gphyses
    variables.collect do |var|
      var.gphys
    end
  end

  def dimensions
    unless @dimensions
      gphys = self.gphyses
      if gphys.length == 0
        @dimensions = []
      else
        gp = gphys[0]
        axes = Array.new
        gp.rank.times do |i|
          axes.push gp.axis(i)
        end
        flag = true
=begin
        (gphys.length-1).times do |i|
          gp = gphys[i+1]
          unless axes.length == gp.rank
            flag = false
            break
          end
          axes.length.times do |j|
            axis = axes[j]
            axis0 = gp.axis(j)
            unless axis.name == axis0.name
              flag = false
              break
            end
          end
          unless flag
            break
          end
        end
=end
        if flag
          @dimensions = axes.collect do |axis|
            pos = axis.pos
            {:name => axis.name, :units => pos.units, :ary => pos.val.to_a}
          end
        else
          @dimensions = []
        end
      end
    end
    return @dimensions
  end

  def index
    dims = region
    if region.length == 0
      dims = Hash.new
      dimensions.each{|dim|
        dims[dim[:name]] = {"min"=>dim[:ary][0], "max"=>dim[:ary][-1]}
      }
    end
    hash = Hash.new
    dims.each do |name,range|
      if ACTION_TYPE[action_type]=="analysis" || (name==x_axis || (DRAW_TYPE[draw_type]!="line"&&name==y_axis))
        hash[name] = range["min"].to_f..range["max"].to_f
      else
        hash[name] = range["min"].to_f
      end
    end
    return hash
  end

  def get_vizshots(basename)
    options = Hash.new
    vars = variables
    nvars = variables.length
    case DRAW_TYPE[draw_type]
    when "line"
      if nvars == 0
        return [false, "at least one variable must be selected"]
      end
      options[:method] = :line
      options.update get_vizshot_options_from_attrs(@@draw_line_attrs)
      nfig = nvars
    when "tone"
      if nvars == 0
        return [false, "at least one variable must be selected"]
      elsif ! (tone || contour)
        return [false, "neither tone or contour are not selected"]
      end
      options[:method] = :tone_cont
      options["transpose"] = true if get_index_from_name(x_axis) > get_index_from_name(y_axis)
      options.update get_vizshot_options_from_attrs(@@draw_tone_attrs)
      nfig = nvars
    when "vector"
      if nvars%2 != 0
        return [false, "number of variable must be multiples of 2"]
      end
      options[:method] = :vector
      options["transpose"] = true if get_index_from_name(x_axis) > get_index_from_name(y_axis)
      options.update get_vizshot_options_from_attrs(@@draw_vector_attrs)
      nfig = nvars/2
    else
      raise "[BUG] figure type is invalid"
    end
    options[:cut] = index
    size = Analysis::DRAW_SIZE[draw_size]
    vizs = Array.new
    nfig.times{|n|
      if Analysis::DRAW_TYPE[draw_type]=="vector"
        options[:variable] = vars[vector_variables_order[n][0]]
        options[:variable2] = vars[vector_variables_order[n][1]]
      else
        options[:variable] = vars[n]
      end
      viz = NumRu::VizShot.new(:iwidth => size[0], :iheight => size[1], :basename => basename)
      viz.set_fig('itr' => draw_projection, 'viewport' => viewport)
      viz.set_tone('tonf' => true )
      viz.plot( options.dup )
      vizs.push viz
    }
    return [true, vizs]
  end




  def diagram_label
    label = DRAW_TYPE[draw_type]
    label += ", #{region[x_axis][:name]}(#{region[x_axis]["min"]}:#{region[x_axis]["max"]})"
    unless DRAW_TYPE[draw_type] == "line"
    label += " vs #{region[y_axis][:name]}(#{region[y_axis]["min"]}:#{region[y_axis]["max"]})"
    end
    tmp = Array.new
    dimensions.each{|dim|
      name = dim[:name]
      next if name==x_axis
      next if DRAW_TYPE[draw_type]!="line" && name==y_axis
      tmp.push "#{name}=#{region[name]["min"]}" if region[name]
    }
    label += " @ "+tmp.join(",") unless tmp.length==0
    return label
  end


  def self.from_vizshot(viz)
    vars, params = parse_vizshot(viz)
    analysis = Analysis.new
    vars.each{|var|
      analysis.variables.push var
    }
    analysis.action_type = ACTION_TYPE.index("draw")
    params.each{|name,val|
      analysis.send(name+"=", val)
    }
    return analysis
  end


  def self.uri_params_from_vizshot(viz)
    vars, params = parse_vizshot(viz)
    ary = Array.new
    vars.each{|var|
      ary.push "variables[#{var.path}]=1"
    }
    ary.push "action_type=draw"
    params.each{|name,val|
      ary += param_to_str("analysis[#{name}]",val)
    }
#    params.collect{|str| str.gsub(/[/,"%5B").gsub(/]/,"%5D")}.join("&")
    URI.escape( ary.join("&") )
  end



  private
  def self.parse_vizshot(viz)
    plot = viz.get_plots
    return nil unless plot.length==1
    plot = plot.first
    plot.each{|k,v| plot[k.to_sym] = v if String === k }

    vars = Array.new
    viz.get_variables.each{|var|
      # return nil unless var.id   # commented out by horinout on 2007-06-28
      vars.push var
    }

    params = Array.new

    cut = plot[:cut]
    if Array === cut
      gphys = GPhys::IO.open(plot[:file], plot[:var])
      ary = cut
      cut = Hash.new
      gphys.rank.times{|i|
        name = gphys.coord(i).name
        cut[name] = ary[i]
      }
    end
    reg = Hash.new
    axes = Array.new
    cut.each{|name,range|
      case range
      when Range
        reg[name] = {"min"=>range.first,"max"=>range.last}
        axes.push name
      when Numeric
        reg[name] = {"min"=>range}
      else
        raise "[BUG]"
      end
    }
    params.push ["region", reg]

    case plot[:method]
    when :line
      draw_type = 0
      unless axes.length==1
        raise "[BUG]"
      end
      twoD = false
    when :tone_cont
      draw_type = 1
      unless axes.length==2
        raise "[BUG]"
      end
      twoD = true
    when :vector
      draw_type = 2
      unless axes.length==2
        raise "[BUG]"
      end
      twoD = true
    else
      raise "[BUG]"
    end
    params.push ["draw_type", draw_type]

    axes = [axes[1],axes[0]] if twoD && plot[:transpose]
    params.push ["x_axis", axes[0]]
    params.push ["y_axis", axes[1]] if twoD
    params.push ["draw_projection", viz.get_itr]
    params.push ["viewport", viz.get_viewport]

    isize = DRAW_SIZE.index(viz.get_size)
    unless isize
      raise "[BUG]"
    end
    params.push ["draw_size", isize]

    case plot[:method]
    when :line
      params += get_params_from_vizshot_plot(@@draw_line_attrs, plot)
    when :tone_cont
      params += get_params_from_vizshot_plot(@@draw_tone_attrs, plot)
    when :vector
      params += get_params_from_vizshot_plot(@@draw_vector_attrs, plot)
      params.push ["vector_variables_order", {0 => {0=>0,1=>1}}]
    end
      return [vars, params]
  end


  def self.param_to_str(name, val)
    ary = Array.new
    case val
    when String
      ary.push "#{name}=#{val}"
    when Numeric
      val = val.to_i if Float === val && val.to_i == val
      ary.push "#{name}=#{val}"
    when TrueClass
      ary.push "#{name}=1"
    when FalseClass
      ary.push "#{name}=0"
    when Array
      ary.push "#{name}=#{val.join(',')}"
    when Hash
      val = val.sort{|a,b|
        if String===a
          if a=="min"
            return -1
          elsif a=="max"
            return 1
          end
        end
        a <=> b
      }
      val.each{|k,v|
        ary += param_to_str("#{name}[#{k}]", v)
      }
    else
      ary.push "#{name}=#{val.inspect}"
    end
    return ary
  end


  def self.get_params_from_vizshot_plot(attrs, plot)
    ary = Array.new
    attrs.each{|attr|
      case attr[4]
      when :vizshot
        name = attr[0].to_sym
      when :ggraph
        name = attr[0]
      else
        next
      end
      val = plot[name]
      ary.push([attr[0],val]) if val
    }
    return ary
  end



  def variables_order(val)
    unless Hash === val
      return nil
    end
    variables_order = Hash.new
    val.each{|k,v|
      if Hash === v
        tmp = Array.new
        v.each{|k1,v1|
          tmp[k1.to_i] = v1.to_i
        }
        variables_order[k.to_i] = tmp
      else
        return nil
      end
    }
    ary = variables_order.values.flatten
    unless ary.max == ary.length-1
      return nil
    end
    ary.length.times{|i|
      unless ary.include?(i)
        return nil
      end
    }
    return variables_order
  end


  def get_vizshot_options_from_attrs(attrs)
    hash = Hash.new
    attrs.each{|attr|
      case attr[4]
      when :vizshot
        name = attr[0].to_sym
      when :ggraph
        name = attr[0]
      else
        next
      end
      val = self.send(attr[0])
      if !attr[3] || (!val.nil? && val!=[] && val!="")
        hash[name] = val
      end
    }
    return hash
  end

  def get_index_from_name(name)
    dimensions.each_with_index{|dim,i|
      if dim[:name] == name
        return i
      end
    }
    return nil
  end



end
